"""
This module handles the creation of webhook subscriptions. It creates
a subscription to receive webhooks for a particular event_type.
"""
from k2connect import exceptions
from k2connect import json_builder
from k2connect import service
from k2connect import validation

WEBHOOK_SUBSCRIPTION_PATH = 'api/v1/webhook_subscriptions'


class WebhookService(service.Service):
    """
    The WebhookService class contains methods for the creation of a webhook
    subscription.
    Example:
        # initialize webhook service
        >>> k2-connect.WebhookService
        >>> k2-connect.create_subscription('buygoods_transaction_reversed',
        >>>................................'https://myapplication/webhooks',
        >>>................................os.getenv('SOME_UNCRACKABLE_SECRET'))

    """

    def __init__(self, base_url):
        """
        :param base_url:
        :type  base_url: str
        """
        super(WebhookService, self).__init__(base_url)

    def create_subscription(self, kwargs):
        """
        Creates a subscription to a webhook service.
        Returns a request response object < class, 'requests.models.Response'>
        :param bearer_token: Access token to be used to make calls to
        the Kopo Kopo API
        :param kwargs: The values constitute all user input.
        :type kwargs: dict
        :return: requests.models.Response
        """
        if 'access_token' not in kwargs:
            raise exceptions.InvalidArgumentError('Access Token not given.')
        if 'event_type' not in kwargs:
            raise exceptions.InvalidArgumentError('Event Type not given.')

        if 'webhook_endpoint' not in kwargs or \
                'scope' not in kwargs:
            raise exceptions.InvalidArgumentError('Invalid arguments for creating a Webhook Subscription.')

        if 'access_token' in kwargs:
            bearer_token = kwargs['access_token']
        if 'event_type' in kwargs:
            event_type = kwargs['event_type']
        if 'webhook_endpoint' in kwargs:
            webhook_endpoint = kwargs['webhook_endpoint']
        if 'scope' in kwargs:
            scope = kwargs['scope']
        if 'scope_reference' in kwargs:
            scope_reference = kwargs['scope_reference']

        # event types
        event_types_to_check = ['b2b_transaction_received', 'buygoods_transaction_received',
                                'buygoods_transaction_reversed', 'm2m_transaction_received',
                                'settlement_transfer_completed', 'customer_created']
        till_scope_event_types = ['b2b_transaction_received', 'buygoods_transaction_received',
                                  'buygoods_transaction_reversed']

        # build subscription url
        subscription_url = self._build_url(WEBHOOK_SUBSCRIPTION_PATH)

        # define headers
        headers = dict(self._headers)

        # validate string arguments
        validation.validate_string_arguments(bearer_token,
                                             event_type,
                                             webhook_endpoint,
                                             scope)

        if 'scope_reference' in kwargs:
            validation.validate_string_arguments(scope_reference)

        headers['Authorization'] = 'Bearer ' + bearer_token + ''

        if not any(check in event_type for check in event_types_to_check):
            raise exceptions.InvalidArgumentError('Event type not recognized by k2-connect')

        if any(check in event_type for check in till_scope_event_types):
            if scope != 'till':
                raise exceptions.InvalidArgumentError('Invalid scope for given event type.')
            if 'scope_reference' not in kwargs or kwargs['scope_reference'] is None:
                raise exceptions.InvalidArgumentError('Scope reference not given.')

        if not any(check in event_type for check in till_scope_event_types):
            scope_reference = None
            if scope != 'company':
                raise exceptions.InvalidArgumentError('Invalid scope for given event type.')
            if 'scope_reference' in kwargs:
                raise exceptions.InvalidArgumentError('Invalid scope reference for given event type.')

        # validate webhook endpoint
        validation.validate_url(webhook_endpoint)

        # define subscription payload
        subscription_payload = json_builder.webhook_subscription(event_type=event_type,
                                                                 webhook_endpoint=webhook_endpoint,
                                                                 scope=scope,
                                                                 scope_reference=scope_reference)

        return self._make_requests(headers=headers,
                                   method='POST',
                                   url=subscription_url,
                                   payload=subscription_payload)

    def webhook_status(self, bearer_token, query_url):
        """
        Returns a JSON object result containing the transaction status.
        :param bearer_token: Access token to be used to make calls to
        the Kopo Kopo API.
        :type bearer_token: str
        :param query_url: URL to which status query is made.
        :type query_url: str
        :return str
        """
        return self._query_transaction_status(bearer_token=bearer_token,
                                              query_url=query_url)



"""
This module handles the creation of setters and python properties
for attributes. These methods set values from decomposed result payloads to
enable access to values in the result JSON payloads.
"""
from k2connect.attribute import attribute


class K2Model:
    """
    The Model class for creating setters and properties based on the
    attributes of different JSON objects
    """

    # attribute(
    #     # create properties and setters for JSON payload generic variables
    #     first_name=None,
    #     middle_name=None,
    #     last_name=None,
    #     amount=None,
    #     currency=None,
    #     status=None,
    #     reference=None,
    #     origination_time=None,
    #     msisdn=None,
    #     # create properties and setters for JSON payloads with errors"
    #     error_code=None,
    #     error_description=None,
    #     # create properties and setters for JSON payloads with links
    #     links_resource=None,
    #     links_self=None,
    #     payment_request=None,
    #     # create properties and setters for PAY JSON payloads
    #     destination=None,
    #     # create properties and setters for JSON payload in Receive MPESA service
    #     payment_result_status=None,
    #     payment_result_id=None,
    #     # create properties and setters for Settlement service JSON payloads
    #     transfer_time=None,
    #     transfer_type=None,
    #     destination_type=None,
    #     transfer_mode=None,
    #     bank=None,
    #     branch=None,
    #     account_number=None,
    #     # properties and setters for b2b transaction
    #     sending_till=None,
    #     # properties and setters for Merchant to Merchant transaction
    #     sending_merchant=None)



"""
"attribute.py" creates simple python properties, getters
and setters functionality for  for attributes passed to the
attribute(key=value) function in a class.
A property name as well as private variable __name is created
and made available of instances of that class

    Example::
        >>>from k2connect.attribute import attribute
        >>>class WinterIsComing:
            ...attribute(name='John Snow')

        >>>instance_of_my_class = WinterIsComing()
        >>>instance_of_my_class.name
        'John Doe'
        >>>instance_of_my_class.name = 'Changing name to : Arya Stark'
        >>>instance_of_my_class.name
        'Changing name to : Arya Stark'
"""

import sys

__all__ = ['attribute', 'readable', 'writable']
__version__ = '3.0'
__author__ = 'Sean Ross'
__credits__ = ['Guido van Rossum', 'Garth Kidd']
__refactor__ = 'Philip Wafula'
__created__ = '10/21/02'
__modified__ = '02/07/19'


def mangle(classname, attribute_name):
    """Mangles name according to python name-mangling
       conventions for private variables"""
    return "_%s__%s" % (classname, attribute_name)


def class_space(class_level=3):
    """Returns the calling class' name and dictionary"""
    frame = sys._getframe(class_level)
    classname = frame.f_code.co_name
    class_dict = frame.f_locals
    return classname, class_dict


# create getting function
def readable(**kwds):
    """Returns one read-only property for each (key,value) pair in kwds"""
    return _attribute(permission='r', **kwds)


# create setting function
def writable(**kwds):
    """Returns one write-only property for each (key,value) pair in kwds"""
    return _attribute(permission='w', **kwds)


# needed because of the way class_space is resolved in _attribute
def attribute(permission='rwd', **kwds):
    """Returns one property for each (key,value) pair in kwds;
       each property provides the specified level of access(permission):
           'r': readable, 'w':writable, 'd':deletable
    """
    return _attribute(permission, **kwds)


# based on code by Guido van Rossum, comp.lang.python 2001-07-31
def _attribute(permission='rwd', **kwds):
    """Returns one property for each (key,value) pair in kwds;
       each property provides the specified level of access(permission):
           'r': readable, 'w':writable, 'd':deletable
    """
    classname, class_dict = class_space()

    def _property(property_attribute_name, property_default):
        property_name, property_attribute_name = property_attribute_name, \
                                                 mangle(classname, property_attribute_name)
        fget, fset, fdel, doc = None, None, None, property_name
        attr_name = property_attribute_name
        if 'r' in permission:
            def fget(self):
                value = property_default
                try:
                    value = getattr(self, attr_name)
                except AttributeError:
                    setattr(self, attr_name, property_default)
                return value
        if 'w' in permission:
            def fset(self, value):
                setattr(self, attr_name, value)
        if 'd' in permission:
            def fdel(self):
                try:
                    delattr(self, attr_name)
                except AttributeError:
                    pass
                # calling fget can restore this attribute, so remove property
                delattr(self.__class__, property_name)
        return property(fget=fget, fset=fset, fdel=fdel, doc=doc)

    for attribute_name, default in kwds.items():
        class_dict[attribute_name] = _property(attribute_name, default)





"""
This module handles OAuth authentication requests using client credentials
i.e. client id and client secret.Once authenticated, the module responds
with a payload containing an access token and its expiry duration. The
access (Bearer) token will be used to authorize the user when requesting
access to k2 resources .
"""
from urllib.parse import urlencode
from k2connect import service
from k2connect import validation

# path for authorization requests
AUTHORIZATION_PATH = 'oauth/token'
REVOKE_PATH = 'oauth/revoke'
INTROSPECT_PATH = 'oauth/introspect'
TOKEN_INFO_PATH = 'oauth/token/info'


class TokenService(service.Service):
    """
    The TokenService class containing methods to request access tokens.
    Example:
        >>>import k2connect
        >>>k2connect.initialize('sample_client_id', 'sample_client_secret', 'https://some_url.com/')
        >>>authenticator = k2-connect.Tokens
        >>>authenticator.request_access_token()
    Get access token and expiry duration
    Example:
        >>>token_response = authenticator.request_access_token()
        >>>access_token = get_access_token(token_service_response)
        >>>expiry_duration = get_token_expiry_duration(token_service_response)
    """

    def __init__(self,
                 base_url,
                 client_id,
                 client_secret):
        """
        Initializes class and inherits all necessary values
        from super class initialization as passed in the library
        initialization function.
        :param base_url: The domain to use in the library.
        :type base_url: str
        :param client_id: Identifier for the k2 user.
        :type client_id: str
        :param client_secret: Secret key for k2 user.
        :type client_secret: str
        """
        validation.validate_string_arguments(base_url,
                                             client_id,
                                             client_secret)

        super(TokenService, self).__init__(base_url)

        self._client_id = client_id
        self._client_secret = client_secret

    def request_access_token(self):
        """
        Returns response object with payload containing access token and
        expiry time.
        """

        # build URL for token request
        url = self._build_url(AUTHORIZATION_PATH)

        # redefine headers for token request
        headers = dict(self._headers)

        # add content-type
        headers['Content-Type'] = 'application/x-www-form-urlencoded'

        # define client credentials payload
        client_credentials_payload = {
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'grant_type': 'client_credentials',
        }

        # url-encode payload
        data = urlencode(client_credentials_payload)

        # request access token and expiry duration
        access_token_request = self._make_requests(data=data, headers=headers, method='POST', url=url)

        return access_token_request

    def revoke_access_token(self, access_token):
        """
        Returns response object with payload containing access token and
        expiry time.
        """

        # build URL for token request
        url = self._build_url(REVOKE_PATH)

        # redefine headers for token request
        headers = dict(self._headers)

        # add content-type
        headers['Content-Type'] = 'application/x-www-form-urlencoded'

        # define client credentials payload
        client_credentials_payload = {
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'token': access_token,
        }

        # url-encode payload
        data = urlencode(client_credentials_payload)

        # request access token and expiry duration
        access_token_request = self._make_requests(data=data, headers=headers, method='POST', url=url)

        return access_token_request

    def introspect_access_token(self, access_token):
        """
        Returns response object with payload containing access token and
        expiry time.
        """

        # build URL for token request
        url = self._build_url(INTROSPECT_PATH)

        # redefine headers for token request
        headers = dict(self._headers)

        # add content-type
        headers['Content-Type'] = 'application/x-www-form-urlencoded'

        # define client credentials payload
        client_credentials_payload = {
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'token': access_token,
        }

        # url-encode payload
        data = urlencode(client_credentials_payload)

        # request access token and expiry duration
        access_token_request = self._make_requests(data=data, headers=headers, method='POST', url=url)

        return access_token_request

    def request_token_info(self, access_token):
        """
        Returns response object with payload containing access token and
        expiry time.
        """

        # build URL for token request
        url = self._build_url(TOKEN_INFO_PATH)

        # redefine headers for token request
        headers = dict(self._headers)

        # add content-type
        headers['Content-Type'] = 'application/x-www-form-urlencoded'
        headers['access_token'] = access_token

        # define client credentials payload
        client_credentials_payload = {
        }

        # url-encode payload
        data = urlencode(client_credentials_payload)

        # request access token and expiry duration
        access_token_request = self._make_requests(data=data, headers=headers, method='POST', url=url)

        return access_token_request

    @staticmethod
    def get_access_token(response):
        """Returns a str containing access token"""
        # get access toke value
        access_token = response.get('access_token')
        return access_token

    @staticmethod
    def get_token_expiry_duration(response):
        """Return duration to expiry of access token"""
        token_expiry_duration = response.json().get('expires_in')
        return token_expiry_duration



"""Handles custom exceptions"""


class K2Error(Exception):
    """Base exception for all errors raised by the K2-Connect library"""

    def __init__(self, error=None):
        if error is None:
            # default error message
            error = "An error has occurred  in the k2-connect library"
        super(K2Error, self).__init__(error)


class InvalidArgumentError(K2Error):
    """The argument passed is invalid, pass message to error description"""



"""
This module handles transfer of funds to pre-approved settlement accounts. It creates
verified settlement bank accounts. Once an account to which funds can be settled is available,
it creates blind and targeted settlement transactions.
Blind transfers are made to the default settlement account defined upon k2 customer acquisition,
targeted transfers are made to a defined destination account.
"""
from k2connect import json_builder, exceptions, service, validation

TRANSFER_PATH = 'api/v1/settlement_transfers'
SETTLEMENT_BANK_ACCOUNTS_PATH = 'api/v1/merchant_bank_accounts'
SETTLEMENT_MOBILE_ACCOUNTS_PATH = 'api/v1/merchant_wallets'


class TransferService(service.Service):
    """
    The TransferService class containing methods for creation of settlement accounts:
    Example:
        # initialize transfer service
        >>> k2connect.TransferService
        >>> k2connect.add_bank_settlement_account('Arya Stark',
        >>>.....................................'45164-IRON BANK',
        >>>.....................................'78491631254523',
        >>>.....................................'564123456987845')
    """

    def __init__(self, base_url):
        """
        Initializes transfer services with the bearer token as an argument.
        This feature allows the developer to refresh the access token
        at any point in their codebase.
        :param bearer_token: Access token to be used to make calls to
        the Kopo Kopo API
        :type  bearer_token: str
        """
        super(TransferService, self).__init__(base_url)

    def add_bank_settlement_account(self, kwargs):
        """
        Creates a verified settlement bank account.
        Returns a request response object < class, 'requests.models.Response'>
        :return: requests.models.Response
        """
        if 'access_token' not in kwargs:
            raise exceptions.InvalidArgumentError('Access Token not given.')

        if 'settlement_method' not in kwargs or \
                'account_name' not in kwargs or \
                'account_number' not in kwargs or \
                'bank_branch_ref' not in kwargs:
            raise exceptions.InvalidArgumentError('Invalid arguments for creating Bank Settlement Account.')

        # iterate through kwargs
        if 'access_token' in kwargs:
            bearer_token = kwargs['access_token']
        if 'settlement_method' in kwargs:
            settlement_method = kwargs['settlement_method']
        if 'account_name' in kwargs:
            account_name = kwargs['account_name']
        if 'account_number' in kwargs:
            account_number = kwargs['account_number']
        if 'bank_branch_ref' in kwargs:
            bank_branch_ref = kwargs['bank_branch_ref']

        # build url
        create_bank_settlement_account_url = self._build_url(SETTLEMENT_BANK_ACCOUNTS_PATH)

        # define headers
        headers = dict(self._headers)

        # validate string arguments
        validation.validate_string_arguments(bearer_token,
                                             settlement_method,
                                             account_name,
                                             bank_branch_ref,
                                             account_number)

        # add authorization to headers
        headers['Authorization'] = 'Bearer ' + bearer_token + ''
        # define create bank settlement account payload
        create_bank_settlement_account_payload = json_builder.bank_settlement_account(settlement_method=settlement_method,
                                                                                      account_name=account_name,
                                                                                      account_number=account_number,
                                                                                      bank_branch_ref=bank_branch_ref)
        return self._make_requests(headers=headers,
                                   method='POST',
                                   url=create_bank_settlement_account_url,
                                   payload=create_bank_settlement_account_payload)

    def add_mobile_wallet_settlement_account(self, kwargs):
        """
        Creates a verified settlement bank account.
        Returns a request response object < class, 'requests.models.Response'>
        :param kwargs: The values constitute all user input.
        :type kwargs: dict
        :return: requests.models.Response
        """
        if 'access_token' not in kwargs:
            raise exceptions.InvalidArgumentError('Access Token not given.')

        if 'first_name' not in kwargs or \
                'last_name' not in kwargs or \
                'phone_number' not in kwargs or \
                'network' not in kwargs:
            raise exceptions.InvalidArgumentError('Invalid arguments for creating Outgoing Pay.')

        # iterate through kwargs
        if 'access_token' in kwargs:
            bearer_token = kwargs['access_token']
        if 'first_name' in kwargs:
            first_name = kwargs['first_name']
        if 'last_name' in kwargs:
            last_name = kwargs['last_name']
        if 'phone_number' in kwargs:
            phone_number = kwargs['phone_number']
        if 'network' in kwargs:
            network = kwargs['network']

        # build url
        create_mobile_wallet_settlement_account_url = self._build_url(SETTLEMENT_MOBILE_ACCOUNTS_PATH)

        # define headers
        headers = dict(self._headers)

        # validate string arguments
        validation.validate_string_arguments(bearer_token,
                                             first_name,
                                             last_name,
                                             phone_number,
                                             network)

        validation.validate_phone_number(phone_number)

        # add authorization to headers
        headers['Authorization'] = 'Bearer ' + bearer_token + ''
        # define create mobile settlement account payload
        create_mobile_settlement_account_payload = json_builder.mobile_settlement_account(first_name=first_name,
                                                                                          last_name=last_name,
                                                                                          phone_number=phone_number,
                                                                                          network=network)
        return self._make_requests(headers=headers,
                                   method='POST',
                                   url=create_mobile_wallet_settlement_account_url,
                                   payload=create_mobile_settlement_account_payload)

    def settle_funds(self, kwargs):
        """
        Creates a transfer from merchant account to a different settlement account.
        Returns a request response object < class, 'requests.models.Response'>
        :param kwargs: The values constitute all user input.
        :type kwargs: dict
        :return: requests.models.Response
        """
        if 'access_token' not in kwargs:
            raise exceptions.InvalidArgumentError('Access Token not given.')
        if 'currency' not in kwargs:
            currency = 'KES'
        if 'destination_reference' not in kwargs:
            destination_reference = ''
        if 'destination_type' not in kwargs:
            destination_type = ''

        if 'callback_url' not in kwargs or \
                'value' not in kwargs:
            raise exceptions.InvalidArgumentError('Invalid arguments for creating Outgoing Pay.')

        # iterate through kwargs
        if 'access_token' in kwargs:
            bearer_token = kwargs['access_token']
        if 'callback_url' in kwargs:
            callback_url = kwargs['callback_url']
        if 'destination_type' in kwargs:
            destination_type = kwargs['destination_type']
        if 'destination_reference' in kwargs:
            destination_reference = kwargs['destination_reference']
        if 'value' in kwargs:
            value = kwargs['value']

        # build settle funds url
        settle_funds_url = self._build_url(TRANSFER_PATH)

        # define headers
        headers = dict(self._headers)

        # check bearer token
        validation.validate_string_arguments(bearer_token,
                                             currency,
                                             value)

        # add authorization to headers
        headers['Authorization'] = 'Bearer ' + bearer_token + ''

        # define amount
        transfer_amount = json_builder.amount(currency=currency,
                                              value=value)

        # create links json object
        transfer_links = json_builder.links(callback_url=callback_url)

        if destination_reference is None and destination_type is None:
            settle_funds_payload = json_builder.transfers(transfer_links=transfer_links,
                                                          transfers_amount=transfer_amount)
        else:
            settle_funds_payload = json_builder.transfers(transfer_links=transfer_links,
                                                          transfers_amount=transfer_amount,
                                                          destination_type=destination_type,
                                                          destination_reference=destination_reference)
        return self._make_requests(headers=headers,
                                   method='POST',
                                   url=settle_funds_url,
                                   payload=settle_funds_payload)

    def transfer_transaction_status(self,
                                    bearer_token,
                                    query_url):
        """
        Returns a JSON object result containing the transaction status.
        :param bearer_token: Access token to be used to make calls to
        the Kopo Kopo API.
        :type bearer_token: str
        :param query_url: URL to which status query is made.
        :type query_url: str
        :return str
        """
        return self._query_transaction_status(bearer_token=bearer_token,
                                              query_url=query_url)

    @staticmethod
    def transfer_transaction_location(response):
        """
        Returns location of the transfers transaction result as returned in the headers of the
        response body.
        :param response: response object from a HTTP request
        :type response: requests.models.Response
        :return str
        """
        return service.k2_requests.get_location(response)



"""
This module handles the processing of response objects from k2-connect. It verifies
the authenticity of a response object by comparing the 'X-KopoKopo-Signature' with
a locally SHA-256 encrypted response body to validate the response.
"""
import hashlib
import hmac
import json

from k2connect import service
from k2connect import exceptions


class ResultProcessor(service.Service):
    """
    The RequestProcessor class containing methods to validate responses
    Example:
        >>>import k2connect
        >>>k2connect.initialize('sample_client_id', 'sample_client_secret', 'https://some_url.com/')
        >>> k2-connect.RequestProcessor
        >>> k2-connect.process(result)
    """
    def __init__(self, base_url, api_secret):
        """
        Initialize processor service with client id and client secret values for
        encryption with sha-256.
        """
        super(ResultProcessor, self).__init__(base_url)
        self._api_secret = api_secret

    def process(self, result):
        """
        Signs result body with a hmac signature. It the compares the resultant signature
        with the X-Kopo-Kopo signature from the headers of the result.
        If the signatures are the same, the result is valid.
        Returns a json formatted str containing content of the HTTP result.
        :param result: A HTTP result object
        :type result: requests.models.Response
        :return: str
        """
        if result is None or result == '':
            raise exceptions.InvalidArgumentError('Response cannot be empty')

        # define result body
        result_body = result.get_data()

        # define X-KopoKopo-Signature
        x_kopo_kopo_signature = result.headers.get('X-KopoKopo-Signature')

        # define hmac signature
        hmac_signature = generate_hmac_signature(bytes(self._api_secret, 'utf-8'), result_body)

        # compare signatures
        if hmac.compare_digest(hmac_signature, x_kopo_kopo_signature) is False:
            print("Hmac: ", hmac)
            print("Signature: ", x_kopo_kopo_signature)
            raise exceptions.InvalidArgumentError('Invalid result passed')
        return json.dumps(result.get_json(), sort_keys=True)


def generate_hmac_signature(api_secret, result_body):
    signature = hmac.new(api_secret, result_body, hashlib.sha256).hexdigest()
    return signature



"""
This module handles the building of JSON objects for HTTPS requests
in the k2-connect library. It serializes python dictionaries to a JSON
formatted strings.
Example:
    >>>import json
    >>>sample_dict = {'key': 'value'}
    >>>print(json.dumps(sample_dict))
    "{"key": "value"}"
"""
import json
from k2connect import exceptions
from k2connect import validation


# serialize python dicts to JSON str
def serializer(dictionary):
    """
    Serializes python dict to json formatted str.
    Returns str.

    :param dictionary:
    :type dictionary: dict
    :return: str
    """
    validation.validate_dictionary_arguments(dictionary)
    return json.dumps(dictionary)


# special function
def metadata(join=None, **kwargs):
    """
    Returns a json formatted str with a maximum of 5 metadata key-value pairs.
    :param join:
    :param kwargs: Metadata information
    :return: str
    """
    if kwargs is None or kwargs == '':
        raise exceptions.InvalidArgumentError('Invalid argument passed. Expects ' / 'key=word/')
    if len(kwargs) >= 5:
        raise exceptions.InvalidArgumentError('Should be less than or 5 metadata objects')
    metadata_object = kwargs
    return metadata_object


def links(callback_url):
    """
    Validates url passed. Returns a str containing a callback_url.
    :param callback_url: Callback URL to post result.
    :type callback_url: str
    :return: str
    """
    # validate string argument
    validation.validate_string_arguments(callback_url)

    # validate url value
    validation.validate_url(callback_url)

    links_object = {'callback_url': callback_url}
    return links_object


def amount(currency, value):
    """
    Returns json formatted str containing currency and value.
    :param currency: Currency of amount
    :type currency: str
    :param value: Value of money
    :type value: str
    :return: str
    """
    # validate str arguments
    validation.validate_string_arguments(currency, value)

    amount_object = {'currency': currency, 'value': value}
    return amount_object


def bank_account(account_name,
                 account_number,
                 settlement_method,
                 bank_branch_ref,
                 **kwargs):
    """
    Returns a json formatted str with the optional value of email and phone.
    Checks if optional values are present, else passes values as null.
    :param account_name: Name as indicated on the bank account name.
    :type account_name: str
    :param account_number: Bank account number.
    :type account_number: str
    :param bank_branch_ref: Identifier identifying the destination bank branch.
    :type bank_branch_ref: str
    :param settlement_method: Whether bank account can settle via EFT or RTS
    :type settlement_method: str
    :param kwargs: Provision for optional 'email' and 'phone' information.
    :type kwargs: kwargs
    :return: str
    """
    # validate string arguments
    validation.validate_string_arguments(account_name,
                                         account_number,
                                         settlement_method,
                                         bank_branch_ref)

    bank_account_object = {'account_name': account_name,
                           'account_number': account_number,
                           'settlement_method': settlement_method,
                           'bank_branch_ref': bank_branch_ref}
    return bank_account_object


def bank_settlement_account(settlement_method,
                            account_name,
                            account_number,
                            bank_branch_ref):
    """
    Returns a json formatted str with bank settlement account information.
    :param settlement_method: EFT or RTS method to transfer funds
    :type settlement_method: str
    :param account_name: The name as indicated on the bank account name
    :type account_name: str
    :param account_number: The bank account number
    :type account_number: str
    :param bank_branch_ref: An identifier identifying the destination bank branch
    :type bank_branch_ref: str
    """
    # validate string arguments
    validation.validate_string_arguments(settlement_method,
                                         account_name,
                                         account_number,
                                         bank_branch_ref)

    bank_settlement_account_object = {'settlement_method': settlement_method,
                                      'account_name': account_name,
                                      'account_number': account_number,
                                      'bank_branch_ref': bank_branch_ref
                                      }
    return bank_settlement_account_object


def mobile_settlement_account(first_name,
                              last_name,
                              phone_number,
                              network):
    """
    Returns a json formatted str with bank settlement account information.
    :param first_name: First name of the recipient.
    :type first_name: str
    :param last_name: Last name of the recipient.
    :type last_name: str
    :param phone_number: The mobile phone number
    :type phone_number: str
    :param network: The bank account number
    :type network: str
    """
    # validate string arguments
    validation.validate_string_arguments(first_name,
                                         last_name,
                                         phone_number,
                                         network)

    mobile_settlement_account_object = {'first_name': first_name,
                                        'last_name': last_name,
                                        'phone_number': phone_number,
                                        'network': network
                                        }
    return mobile_settlement_account_object


def mobile_wallet(first_name,
                  last_name,
                  phone_number,
                  network,
                  **kwargs):
    """
    Returns a json formatted str with the optional value of email.
    Checks if optional value is present, else passes values as null.

    :param first_name: First name of the recipient.
    :type first_name: str
    :param last_name: Last name of the recipient.
    :type last_name: str
    :param phone_number: Phone number of recipient.
    :type phone_number: str
    :param network: The mobile network to which the phone number belongs.
    :type network: str
    :param kwargs: Provision for optional 'email' value
    :type kwargs: kwargs
    :return:
    """
    # validate string arguments
    validation.validate_string_arguments(first_name,
                                         last_name,
                                         phone_number,
                                         network)

    if 'email' not in kwargs:
        email = 'Null'
    else:
        email = kwargs['email']

    mobile_wallet_object = {'first_name': first_name,
                            'last_name': last_name,
                            'phone_number': phone_number,
                            'network': network,
                            'email': email
                            }
    return mobile_wallet_object


def pay_recipient(recipient_type, recipient):
    """
    Returns json formatted str containing pay recipient object.
    :param recipient_type: Type of recipient eg. mobile wallet
    or bank account
    :type recipient_type: str
    :param recipient: JSON formatted str containing details of the
    recipient
    :type recipient: str
    :return: str
    """
    # validate string arguments
    validation.validate_string_arguments(*recipient, recipient_type)

    recipient_object = {'type': recipient_type, 'pay_recipient': recipient}
    return recipient_object


def till_pay_recipient(till_name, till_number):
    """
    Returns a json formatted str with the optional value of email.
    Checks if optional value is present, else passes values as null.

    :param till_name: First name of the recipient.
    :type till_name: str
    :param till_number: Last name of the recipient.
    :type till_number: str
    :return:
    """
    # validate string arguments
    validation.validate_string_arguments(till_name,
                                         till_number)

    till_pay_recipient_object = {'till_name': till_name,
                            'till_number': till_number
                            }
    return till_pay_recipient_object


def kopo_kopo_merchant_pay_recipient(alias_name, till_number):
    """
    Returns a json formatted str with the optional value of email.
    Checks if optional value is present, else passes values as null.

    :param alias_name: First name of the recipient.
    :type alias_name: str
    :param till_number: Last name of the recipient.
    :type till_number: str
    :return:
    """
    # validate string arguments
    validation.validate_string_arguments(alias_name, till_number)

    k2_merchant_pay_recipient_object = {'alias_name': alias_name,
                            'till_number': till_number
                            }
    return k2_merchant_pay_recipient_object


def subscriber(first_name,
               last_name,
               phone,
               email,
               **kwargs):
    """
    Returns JSON formatted str containing subscriber information
    :param first_name: First name of the subscriber
    :type first_name: str
    :param last_name: Last name of the subscriber
    :type last_name: str
    :param phone: Phone number of the subscriber from which the
    pay will be made
    :type phone: str
    :param email: Email of the subscriber
    :type email: str
    :param kwargs: Provision for optional 'email' information.
    :type kwargs: str
    :return: str
    """
    # validate string arguments
    validation.validate_string_arguments(first_name,
                                         last_name,
                                         phone)
    if email != "Null":
        validation.validate_email(email)

    # if 'email' not in kwargs:
    #     email = 'Null'
    # else:
    #     email = kwargs['email']

    subscriber_object = {'first_name': first_name,
                         'last_name': last_name,
                         'phone_number': phone,
                         'email': email}

    return subscriber_object


def mpesa_payment(mpesa_links,
                  mpesa_payment_amount,
                  mpesa_payment_subscriber,
                  payment_channel,
                  till_number,
                  **kwargs):
    """
    Returns JSON formatted str with optional 'metadata' JSON str object.
    :param mpesa_links: A JSON object containing the call back URL
    :type mpesa_links: str
    :param payment_channel: Payment channel e.g MPESA
    :type payment_channel: str
    :param till_number: Till to which the pay is made.
    :type till_number: str
    :param mpesa_payment_subscriber: Subscriber JSON object.
    :type mpesa_payment_subscriber str
    :param mpesa_payment_amount: Amount JSON object.
    :type mpesa_payment_amount str
    :param kwargs: Optional JSON object containing a maximum of 5 key
    value pairs.
    :return: str
    """

    # validate string arguments
    validation.validate_string_arguments(*mpesa_links,
                                         *mpesa_payment_amount,
                                         *mpesa_payment_subscriber,
                                         *payment_channel,
                                         *till_number)

    if 'metadata' not in kwargs:
        mpesa_payment_metadata = {}
    else:
        mpesa_payment_metadata = kwargs['metadata']

    mpesa_payment_object = {'payment_channel': payment_channel,
                            'till_number': till_number,
                            'subscriber': mpesa_payment_subscriber,
                            'amount': mpesa_payment_amount,
                            'metadata': mpesa_payment_metadata,
                            '_links': mpesa_links
                            }

    return mpesa_payment_object


def webhook_subscription(event_type,
                         webhook_endpoint,
                         scope,
                         scope_reference=None):
    """
    Returns JSON formatted str containing webhook subscription information
    :param event_type:The type of event subscribed to.
    :type event_type:  str
    :param webhook_endpoint: The HTTP end point to send the webhook.
    :type webhook_endpoint: str
    :param scope: A string that will be used to specify whether account is at Till or Company level.
    :type scope: str
    :param scope_reference: A string that represents the account number (eg MPESA till number).
    :type scope_reference: str
    :return: str
    """

    # validate string arguments
    validation.validate_string_arguments(event_type,
                                         webhook_endpoint,
                                         scope)
    if scope_reference is not None:
        validation.validate_string_arguments(scope_reference)

    webhook_subscription_object = {'event_type': event_type,
                                   'url': webhook_endpoint,
                                   'scope': scope,
                                   'scope_reference': scope_reference
                                   }
    return webhook_subscription_object


def pay(destination_reference,
        destination_type,
        payment_amount,
        payment_links,
        payment_metadata):
    """
    Return JSON formatted str containing information about a transfer.
    :param destination_reference: reference for the pay_recipient account.
    :type destination_reference : str
    :param destination_type: Differentiate between mobile and bank account type for recipient
    :type destination_type : str
    :param payment_amount: A JSON formatted str containing the currency
    and the amount to be transferred.
    :type payment_amount: str
    :param payment_metadata: A JSON formatted str with a maximum of 5
    key-value pairs.
    :type payment_metadata: str
    :param payment_links:A JSON formatted str containing a call back URL.
    :type payment_links: str
    :return: str
    """

    # validate string arguments
    validation.validate_string_arguments(*destination_reference,
                                         *destination_type,
                                         *payment_amount,
                                         *payment_links,
                                         *payment_metadata)

    payment_json_object = {
        "destination_reference": destination_reference,
        "destination_type": destination_type,
        "amount": payment_amount,
        "metadata": payment_metadata,
        "_links": payment_links
    }
    return payment_json_object


def transfers(transfer_links, transfers_amount,
              **kwargs):
    """
    Returns JSON formatted containing information about a transfer.
    :param transfers_amount: Amount to be transferred.
    :type transfers_amount: str
    :param transfer_links: Links containing Callback URL.
    :type transfer_links: str
    :param kwargs: Provision for optional 'destination' value.
    :type kwargs: str
    :return: str
    """
    # validate string arguments
    validation.validate_string_arguments(*transfer_links, *transfers_amount)

    if 'destination_type' not in kwargs:
        destination_type = None
    else:
        destination_type = kwargs['destination_type']

    if 'destination_reference' not in kwargs:
        destination_reference = None
    else:
        destination_reference = kwargs['destination_reference']

    transfers_object = {'amount': transfers_amount,
                        'destination_reference': destination_reference,
                        'destination_type': destination_type,
                        '_links': transfer_links
                        }
    return transfers_object



"""
This module handles HTTP requests for the k2-connect library.
It defines all base requirements for requests made by different
k2-connect services. Furthermore, it handles the access of specific
resources contained in responses of HTTP requests.
"""
import requests
from urllib.parse import urlparse
from k2connect import exceptions
from k2connect import validation


class Requests:
    """
    Performs k2-connect HTTP requests.
    Retrieves values found in k2-connect response objects.
    """

    def __init__(self):
        """
        Initializes function arguments and other variables necessary
        for class and its subclasses.

        :param bearer_token: Access token to be used to make calls to the Kopo Kopo API
        :type  bearer_token: str
        """
        self._headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }

    @staticmethod
    def __get_request(headers, params, url):
        """
        Returns http response.

        :param url: URL to which GET request is sent.
        :type url: str
        :param headers: Headers for GET requests
        :type headers: dict
        :param params: Dictionary, list of tuples or bytes to send
        in the body of the :class:`Request`.
        :type params:
        :return: 'requests.models.Response'
        """
        response = requests.get(
            headers=headers,
            params=params,
            url=url)
        return response

    @staticmethod
    def __post_request(headers, url, payload, data):
        """
        Returns http response.

        :param url: URL to which POST request is sent.
        :type url: str
        :param headers: Headers for POST requests.
        :type headers: str (json)
        :param payload: Payload for POST request
        :type payload: str (json)
        :param data: Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
        :return: 'requests.models.Response'
        """

        response = requests.post(
            headers=headers,
            json=payload,
            data=data,
            url=url,
        )

        return response

    def _make_requests(self,
                       headers,
                       method,
                       url,
                       data=None,
                       payload=None,
                       params=None):
        """
        Returns JSON payload.
        :param headers: Headers for HTTP request
        :type headers: str (JSON)
        :param method: Method for HTTP request (GET, POST)
        :type method str
        :param url: URL to which HTTP request is sent
        :type url: str
        :param data: Dictionary, list of tuples or bytes to send
        in the body of the :class:`Request`.
        :param payload: Payload for HTTP request
        :type payload: str (JSON)
        :param params: Query parameters for a HTTP request
        :type params: str (JSON)
        :return: str (JSON)
        """
        if validation.validate_url(url) is True:
            if method == 'GET':
                response = self.__get_request(url=url,
                                              headers=headers,
                                              params=params)
            elif method == 'POST':
                response = self.__post_request(url=url,
                                               headers=headers,
                                               payload=payload,
                                               data=data)
            else:
                raise exceptions.InvalidArgumentError('Method not recognized by k2-connect')
            # define status code to check
            status_code = response.status_code

            if 200 <= status_code <= 300:
                # FIXME: This is a HACK.
                if urlparse(url).path == '/oauth/token' or method == 'GET':
                    return response.json()
                response_location = response.headers.get('location')
                return response_location
            return response.reason
        return exceptions.K2Error

    def _query_transaction_status(self,
                                  bearer_token,
                                  query_url):
        """
        Returns a JSON object result containing the transaction status.
        :param bearer_token: Access token to be used to make calls to
        the Kopo Kopo API
        :type bearer_token: str
        :param query_url: URL to which status query is made.
        :type query_url: str
        :return: str
        """
        # define headers
        _headers = dict(self._headers)

        # check bearer token
        validation.validate_string_arguments(bearer_token, query_url)
        # add bearer token
        _headers['Authorization'] = 'Bearer ' + bearer_token + ''

        # validate url
        validation.validate_url(query_url)

        return self._make_requests(url=query_url,
                                   method='GET',
                                   headers=_headers)


def get_location(response):
    """
    Returns location of a transaction as returned in the header of a response object.
    :param response: response object succeeding a successful post request.
    :type response: requests.models.Response
    :return: str
    """
    resource_location = response.headers('location')
    return resource_location



"""
This module handles the initialization of connections to the k2-connect
library's services.It defines all functions a k2-connect service class
should have. It is the base services class on top of which k2-connect
services are built.
"""
from k2connect import k2_requests
from k2connect import validation


class Service(k2_requests.Requests):
    """
    The Base service :class: Service  containing all service class
    functionality.
    Used to create custom k2-connect service classes.\
    Example:
        class PayService(Service):
            ...
    """
    def __init__(self,
                 base_url):
        """
        Initializes function arguments and creates protected variables
        accessible to all subclasses of the Service class.

        :param base_url: The domain to use in the library
        :type base_url: str
        """
        super(Service, self).__init__()

        # validate base url
        if validation.validate_url(base_url) is True:
            self._base_url = base_url

    def _build_url(self, url_path):
        """
        Returns complete URL with path for specific services appended to the base URL
        Example:
            Service(client_id, client_secret, 'https://api.kopokopo.com/')
            return: https://api.kopokopo.com/url_path

        :param url_path: path for specific k2-connect services
        :type url_path: str
        :return str
        """
        return self._base_url + url_path
